@@ -4,6 +4,7 @@ Tests for bitarray.util module
 from __future__ import absolute_import
 
 import os
+import sys
 import array
 import base64
 import binascii
@@ -20,7 +21,7 @@ from collections import Counter
 from bitarray import (bitarray, frozenbitarray, decodetree, bits2bytes,
                       _set_default_endian)
 from bitarray.test_bitarray import (Util, skipIf, is_pypy, urandom_2,
-                                    SYSINFO, DEBUG, WHITESPACE)
+                                    PTRSIZE, DEBUG, WHITESPACE)
 
 from bitarray.util import (
     zeros, ones, urandom, pprint, strip, count_n,
@@ -34,7 +35,9 @@ from bitarray.util import (
 )
 
 if DEBUG:
-    from bitarray._util import _sc_rts, _SEGSIZE  # type: ignore
+    from bitarray._util import (  # type: ignore
+        _count_from_word, _read_n, _write_n, _sc_rts, _SEGSIZE,
+    )
     SEGBITS = 8 * _SEGSIZE
 else:
     SEGBITS = None
@@ -297,7 +300,7 @@ class CountN_Tests(unittest.TestCase, Util):
         self.assertRaises(TypeError, count_n, a, 7.0)
         self.assertRaises(ValueError, count_n, a, 0, 2)
         self.assertRaisesMessage(ValueError, "n = 1 larger than bitarray "
-                                 "size (len(a) = 0)", count_n, a, 1)
+                                 "length 0", count_n, a, 1)
 
     def test_simple(self):
         a = bitarray('111110111110111110111110011110111110111110111000')
@@ -306,21 +309,10 @@ class CountN_Tests(unittest.TestCase, Util):
         self.assertEqual(a.count(), 37)
         self.assertEqual(a.count(0), 11)
 
+        self.assertEqual(count_n(a, 0), 0)
         self.assertEqual(count_n(a, 0, 0), 0)
         self.assertEqual(count_n(a, 2, 0), 12)
         self.assertEqual(count_n(a, 10, 0), 47)
-        # n < 0
-        self.assertRaisesMessage(ValueError, "non-negative integer expected",
-                                 count_n, a, -1, 0)
-        # n > len(a)
-        self.assertRaisesMessage(ValueError, "n = 49 larger than bitarray "
-                                 "size (len(a) = 48)", count_n, a, 49, 0)
-        # n > a.count(0)
-        self.assertRaisesMessage(ValueError, "n = 12 exceeds total count "
-                                 "(a.count(0) = 11)",
-                                 count_n, a, 12, 0)
-
-        self.assertEqual(count_n(a, 0), 0)
         self.assertEqual(count_n(a, 20), 23)
         self.assertEqual(count_n(a, 20, 1), 23)
         self.assertEqual(count_n(a, 37), 45)
@@ -329,10 +321,13 @@ class CountN_Tests(unittest.TestCase, Util):
                                  count_n, a, -1)
         # n > len(a)
         self.assertRaisesMessage(ValueError, "n = 49 larger than bitarray "
-                                 "size (len(a) = 48)", count_n, a, 49)
+                                 "length 48", count_n, a, 49)
+        # n > a.count(0)
+        self.assertRaisesMessage(ValueError, "n = 12 exceeds total count "
+                                 "(a.count(0) = 11)", count_n, a, 12, 0)
         # n > a.count(1)
         self.assertRaisesMessage(ValueError, "n = 38 exceeds total count "
-                                 "(a.count(1) = 37)", count_n, a, 38)
+                                 "(a.count(1) = 37)", count_n, a, 38, 1)
 
         for v in 0, 1:
             for n in range(a.count(v) + 1):
@@ -342,7 +337,7 @@ class CountN_Tests(unittest.TestCase, Util):
                 self.assertEqual(i, self.count_n(a if v else ~a, n))
         self.assertEQUAL(a, b)
 
-    def test_frozen(self):
+    def test_frozenbitarray(self):
         a = frozenbitarray('001111101111101111101111100111100')
         self.assertEqual(len(a), 33)
         self.assertEqual(a.count(), 24)
@@ -356,7 +351,7 @@ class CountN_Tests(unittest.TestCase, Util):
             self.check_result(a, n, count_n(a, n))
 
     def test_ones(self):
-        n = randint(1, 100000)
+        n = randint(1, 100_000)
         a = ones(n)
         self.assertEqual(count_n(a, n), n)
         self.assertRaises(ValueError, count_n, a, 1, 0)
@@ -366,7 +361,7 @@ class CountN_Tests(unittest.TestCase, Util):
             self.assertEqual(count_n(a, i), i)
 
     def test_one_set(self):
-        n = randint(1, 100000)
+        n = randint(1, 100_000)
         a = zeros(n)
         self.assertEqual(count_n(a, 0), 0)
         self.assertRaises(ValueError, count_n, a, 1)
@@ -381,17 +376,16 @@ class CountN_Tests(unittest.TestCase, Util):
         for N in range(1, 1000):
             a = zeros(N)
             a[-1] = 1
-            self.assertEqual(a.count(), 1)
             self.assertEqual(count_n(a, 1), N)
             if N == 1:
-                msg = "n = 2 larger than bitarray size (len(a) = 1)"
+                msg = "n = 2 larger than bitarray length 1"
             else:
                 msg = "n = 2 exceeds total count (a.count(1) = 1)"
             self.assertRaisesMessage(ValueError, msg, count_n, a, 2)
 
     def test_large(self):
         for _ in range(100):
-            N = randint(100000, 250000)
+            N = randint(100_000, 250_000)
             a = bitarray(N)
             v = getrandbits(1)
             a.setall(not v)
@@ -409,15 +403,6 @@ class CountN_Tests(unittest.TestCase, Util):
                 i = count_n(a, n, v)
                 self.check_result(a, n, i, v)
 
-    def test_random(self):
-        for a in self.randombitarrays():
-            v = getrandbits(1)
-            n = a.count(v) // 2
-            i = count_n(a, n, v)
-            self.check_result(a, n, i, v)
-            # n = 0 -> count_n always 0
-            self.assertEqual(count_n(a, 0, v), 0)
-
 # ---------------------------------------------------------------------------
 
 class BitwiseCountTests(unittest.TestCase, Util):
@@ -694,10 +679,13 @@ class ByteswapTests(unittest.TestCase, Util):
         byteswap(a)
         self.assertEqual(a, bitarray("01010101 11110000"))
 
-        a = bitarray("11110000 1111")
+        a = bitarray("01111000 1001")
+        b = a.copy()
+        a.tobytes()  # clear padbits
         byteswap(a)
+        self.assertEqual(a, bitarray("10010000 0111"))
         byteswap(a)
-        self.assertEqual(a, bitarray("11110000 1111"))
+        self.assertEqual(a, b)
 
     def test_basic_array(self):
         r = os.urandom(64)
@@ -1217,7 +1205,7 @@ class BaseTests(unittest.TestCase, Util):
         for m, n, alphabet in self.alphabets:
             for i in range(256):
                 c = chr(i)
-                if c in alphabet or c in WHITESPACE:
+                if c in alphabet or c.isspace():
                     continue
                 if n == 16 and c in "ABCDEF":
                     continue
@@ -1291,7 +1279,7 @@ class SC_Tests(unittest.TestCase, Util):
                                      bytearray([0x01, 0x10, c]))
 
     def test_decode_header_overflow(self):
-        nbytes = SYSINFO[1]
+        nbytes = PTRSIZE
         self.assertRaisesMessage(
             OverflowError,
             "sizeof(Py_ssize_t) = %d: cannot read 9 bytes" % nbytes,
@@ -1300,7 +1288,7 @@ class SC_Tests(unittest.TestCase, Util):
         self.assertRaisesMessage(
             ValueError,
             "read %d bytes got negative value: -1" % nbytes,
-            sc_decode, bytes(bytearray([nbytes] + nbytes * [0xff])))
+            sc_decode, bytearray([nbytes] + nbytes * [0xff]))
 
         if nbytes == 4:
             self.assertRaisesMessage(
@@ -1332,27 +1320,22 @@ class SC_Tests(unittest.TestCase, Util):
             ValueError, "decode error (n=3): 32768 >= 32768",
             sc_decode, b"\x02\x00\x80\xc3\x01\x00\x80\x00\0")
 
-        if SYSINFO[1] == 4:
-            msg = "read 4 bytes got negative value: -2147483648"
-        else:
-            msg = "decode error (n=4): 2147483648 >= 16"
+        msg = {4: "read 4 bytes got negative value: -2147483648",
+               8: "decode error (n=4): 2147483648 >= 16"}
         self.assertRaisesMessage(
-            ValueError, msg,
+            ValueError, msg[PTRSIZE],
             sc_decode, b"\x01\x10\xc4\x01\x00\x00\x00\x80\0")
 
-        if SYSINFO[1] == 4:
-            msg = "read 4 bytes got negative value: -1"
-        else:
-            msg = "decode error (n=4): 4294967295 >= 16"
+        msg = {4: "read 4 bytes got negative value: -1",
+               8: "decode error (n=4): 4294967295 >= 16"}
         self.assertRaisesMessage(
-            ValueError, msg,
+            ValueError, msg[PTRSIZE],
             sc_decode, b"\x01\x10\xc4\x01\xff\xff\xff\xff\0")
 
     def test_decode_end_of_stream(self):
         for stream in [b'', b'\x00', b'\x01', b'\x02\x77',
                        b'\x01\x04\x01', b'\x01\x04\xa1', b'\x01\x04\xa0']:
-            self.assertRaisesMessage(ValueError, "unexpected end of stream",
-                                     sc_decode, stream)
+            self.assertRaises(StopIteration, sc_decode, stream)
 
     def test_decode_types(self):
         blob = b'\x11\x03\x01\x20\0'
@@ -1378,7 +1361,7 @@ class SC_Tests(unittest.TestCase, Util):
         for b in [
                 # raw:
                 b'\x11\x03\x01\x20\0',    # this is what sc_encode gives us
-                b'\x11\x03\x01\x2f\0',    # some pad bits are 1
+                b'\x11\x03\x01\x3f\0',    # but we can set the pad bits to 1
                 # sparse:
                 b'\x11\x03\xa1\x02\0',                  # using block type 1
                 b'\x11\x03\xc2\x01\x02\x00\0',          # using block type 2
@@ -1408,7 +1391,7 @@ class SC_Tests(unittest.TestCase, Util):
             b.append(0)  # stop
 
             self.assertEqual(sc_decode(b), a)
-            self.assertEqual(sc_encode(a), bytes(b))
+            self.assertEqual(sc_encode(a), b)
 
     def test_decode_random_bytes(self):
         # ensure random input doesn't crash the decoder
@@ -1417,9 +1400,8 @@ class SC_Tests(unittest.TestCase, Util):
             b = b'\x02\x00\x04' + os.urandom(n)
             try:
                 a = sc_decode(b)
-            except ValueError as e:
-                if e != 'unexpected end of stream':
-                    continue
+            except (StopIteration, ValueError):
+                continue
             self.assertEqual(len(a), 1024)
             self.assertEqual(a.endian, 'little')
 
@@ -1438,7 +1420,7 @@ class SC_Tests(unittest.TestCase, Util):
         b = sc_decode(i)
         self.assertTrue(a == b == c)
         self.assertTrue(a.endian == b.endian == c.endian)
-        self.assertEqual(bytes(i), b'')
+        self.assertEqual(list(i), [])
 
     def test_encode_zeros(self):
         for i in range(18):
@@ -1457,11 +1439,21 @@ class SC_Tests(unittest.TestCase, Util):
             self.assertEqual(m, len(sc_encode(a)))
             self.round_trip(a)
 
-        a = zeros(1 << 25, 'big')
-        a[0] = 1
+    def test_encode_type4(self):
+        a = zeros(1 << 26, 'big')
         self.assertEqual(
             sc_encode(a),
-            b'\x14\x00\x00\x00\x02\xc4\x01\x00\x00\x00\x00\x00')
+            b'\x14\x00\x00\x00\x04\0')
+
+        a[0xccbbaa] = 1
+        self.assertEqual(
+            sc_encode(a),
+            b'\x14\x00\x00\x00\x04\xc4\x01\xaa\xbb\xcc\x00\0')
+
+        a[0x3ffeedd] = 1
+        self.assertEqual(
+            sc_encode(a),
+            b'\x14\x00\x00\x00\x04\xc4\x02\xaa\xbb\xcc\x00\xdd\xee\xff\x03\0')
 
     def test_encode_ones(self):
         for _ in range(50):
@@ -1470,9 +1462,10 @@ class SC_Tests(unittest.TestCase, Util):
             m = 2                                # head byte and stop byte
             m += bits2bytes(nbits.bit_length())  # size bytes
             nbytes = bits2bytes(nbits)
-            m += (nbytes // 32 + 127) // 128  # number of blocks (head bytes)
-            m += bool(nbytes % 32)            # block type 0 range(1, 32)
             m += nbytes                       # actual raw bytes
+            # number of head bytes, all of block type 0:
+            m += bool(nbytes % 32)            # number in 0x01 .. 0x1f
+            m += (nbytes // 32 + 127) // 128  # number in 0x20 .. 0xbf
             self.assertEqual(m, len(sc_encode(a)))
             self.round_trip(a)
 
@@ -1485,25 +1478,22 @@ class SC_Tests(unittest.TestCase, Util):
                 a &= urandom(n, endian)
                 self.round_trip(a)
 
-# ---------------------------------------------------------------------------
+# -----------------  _sc_rts()   (running totals debug test)  ---------------
 
 @skipIf(not DEBUG)
 class RTS_Tests(unittest.TestCase):
 
-    # Internal functionality exposed for the purpose of testing.
-    # This class will only be part of the test suite in debug mode.
-
     def test_segsize(self):
         self.assertIsInstance(_SEGSIZE, int)
         self.assertTrue(_SEGSIZE in [8, 16, 32])
 
-    def test_rts_empty(self):
+    def test_empty(self):
         rts = _sc_rts(bitarray())
         self.assertEqual(len(rts), 1)
         self.assertEqual(rts, [0])
 
     @skipIf(SEGBITS != 256)
-    def test_rts_example(self):
+    def test_example(self):
         # see example before sc_calc_rts() in _util.c
         a = zeros(987)
         a[:5] = a[512:515] = a[768:772] = 1
@@ -1513,26 +1503,28 @@ class RTS_Tests(unittest.TestCase):
         self.assertEqual(len(rts), 5)
         self.assertEqual(rts, [0, 5, 5, 8, 12])
 
-    def test_rts_ones(self):
-        for _ in range(20):
-            n = randrange(10000)
+    @staticmethod
+    def nseg(a):  # number of segments
+        return (a.nbytes + _SEGSIZE - 1) // _SEGSIZE
+
+    def test_ones(self):
+        for n in range(1000):
             a = ones(n)
             rts = _sc_rts(a)
-            self.assertEqual(len(rts), (n + SEGBITS - 1) // SEGBITS + 1)
+            self.assertEqual(len(rts), self.nseg(a) + 1)
             self.assertEqual(rts[0], 0)
             self.assertEqual(rts[-1], n)
             for i, v in enumerate(rts):
                 self.assertEqual(v, min(SEGBITS * i, n))
 
-    def test_rts_random(self):
-        for _ in range(20):
-            n = randrange(10000)
-            a = urandom(n)
+    def test_random(self):
+        for _ in range(200):
+            a = urandom_2(randrange(10000))
             rts = _sc_rts(a)
-            self.assertEqual(len(rts), (n + SEGBITS - 1) // SEGBITS + 1)
+            self.assertEqual(len(rts), self.nseg(a) + 1)
             self.assertEqual(rts[0], 0)
             self.assertEqual(rts[-1], a.count())
-            for i in range(len(rts) - 1):
+            for i in range(self.nseg(a)):
                 seg_pop = a.count(1, SEGBITS * i, SEGBITS * (i + 1))
                 self.assertEqual(rts[i + 1] - rts[i], seg_pop)
 
@@ -1625,15 +1617,15 @@ class VLFTests(unittest.TestCase, Util):
             self.assertEqual(vl_decode(stream), a)
 
     def test_decode_errors(self):
-        # decode empty bits
-        self.assertRaises(ValueError, vl_decode, b'')
+        # decode empty bytes
+        self.assertRaises(StopIteration, vl_decode, b'')
         # invalid number of padding bits
         for s in b'\x50', b'\x60', b'\x70':
             self.assertRaises(ValueError, vl_decode, s)
         self.assertRaises(ValueError, vl_decode, b'\xf0')
         # high bit set, but no terminating byte
         for s in b'\x80', b'\x80\x80':
-            self.assertRaises(ValueError, vl_decode, s)
+            self.assertRaises(StopIteration, vl_decode, s)
         # decode list with out of range items
         for i in -1, 256:
             self.assertRaises(ValueError, vl_decode, [i])
@@ -1663,6 +1655,7 @@ class VLFTests(unittest.TestCase, Util):
         c = a.copy()
         s = vl_encode(a)
         b = vl_decode(s)
+        self.check_obj(b)
         self.assertTrue(a == b == c)
         LEN_PAD_BITS = 3
         self.assertEqual(len(s), (len(a) + LEN_PAD_BITS + 6) // 7)
@@ -1671,13 +1664,10 @@ class VLFTests(unittest.TestCase, Util):
         padding = (head & 0x70) >> 4
         self.assertEqual(len(a) + padding, 7 * len(s) - LEN_PAD_BITS)
 
-    def test_range(self):
-        for n in range(500):
-            self.round_trip(urandom(n))
-
     def test_large(self):
-        a = urandom(randint(50000, 100000))
-        self.round_trip(a)
+        for _ in range(10):
+            a = urandom(randrange(100_000))
+            self.round_trip(a)
 
     def test_random(self):
         for a in self.randombitarrays():
@@ -2042,22 +2032,21 @@ class SerializationTests(unittest.TestCase, Util):
             self.assertRaisesMessage(ValueError, msg, deserialize, b)
 
         for i in range(256):
-            b = bytes(bytearray([i]))
+            b = bytearray([i])
             if i == 0 or i == 16:
                 self.assertEqual(deserialize(b), bitarray())
             else:
                 self.assertRaises(ValueError, deserialize, b)
                 check_msg(b)
 
-            b += b'\0'
+            b.append(0)
             if i < 32 and i % 16 < 8:
                 self.assertEqual(deserialize(b), zeros(8 - i % 8))
             else:
                 self.assertRaises(ValueError, deserialize, b)
                 check_msg(b)
 
-    def test_bits_ignored(self):
-        # the unused padding bits (with the last bytes) are ignored
+    def test_padbits_ignored(self):
         for blob, endian in [
                 (b'\x07\x01', 'little'),
                 (b'\x07\x03', 'little'),
@@ -2286,7 +2275,7 @@ class CanonicalHuffmanTests(unittest.TestCase, Util):
 
             maxbits = 1 << i
             count[i] = maxbits
-            if i == 31 and SYSINFO[1] == 4:
+            if i == 31 and PTRSIZE == 4:
                 self.assertRaises(OverflowError,
                                   canonical_decode, a, count, [])
                 continue
@@ -2492,6 +2481,88 @@ class CanonicalHuffmanTests(unittest.TestCase, Util):
             freq = {i: random() for i in range(n)}
             self.check_code(*canonical_huffman(freq))
 
+# ------------------------- Internal debug tests ----------------------------
+
+@skipIf(not DEBUG)
+class CountFromWord_Tests(unittest.TestCase, Util):
+
+    def test_ones_zeros_empty(self):
+        for _ in range(1000):
+            n = randrange(1024)
+            a = ones(n)
+            i = randrange(16)
+            self.assertEqual(_count_from_word(a, i), max(0, n - i * 64))
+            a.setall(0)
+            self.assertEqual(_count_from_word(a, i), 0)
+            a.clear()
+            self.assertEqual(_count_from_word(a, i), 0)
+
+    def test_random(self):
+        for _ in range(1000):
+            n = randrange(1024)
+            a = urandom_2(n)
+            i = randrange(16)
+            res = _count_from_word(a, i)
+            self.assertEqual(res, a[64 * i:].count())
+
+
+@skipIf(not DEBUG)
+class ReadN_WriteN_Tests(unittest.TestCase, Util):
+
+    # Regardless of machine byte-order, read_n() and write_n() use
+    # little endian byte-order.
+
+    def test_explicit(self):
+        for blob, x in [(b"", 0),
+                        (b"\x00", 0),
+                        (b"\x01", 1),
+                        (b"\xff", 255),
+                        (b"\xff\x00", 255),
+                        (b"\xaa\xbb\xcc", 0xccbbaa)]:
+            n = len(blob)
+            self.assertEqual(_read_n(iter(blob), n), x)
+            self.assertEqual(_write_n(n, x), blob)
+
+    def test_zeros(self):
+        for n in range(PTRSIZE):
+            blob = n * b"\x00"
+            self.assertEqual(_read_n(iter(blob), n), 0)
+            self.assertEqual(_write_n(n, 0), blob)
+
+    def test_max(self):
+        blob = (PTRSIZE - 1) * b"\xff" + b"\x7f"
+        self.assertEqual(_read_n(iter(blob), PTRSIZE), sys.maxsize)
+        self.assertEqual(_write_n(PTRSIZE, sys.maxsize), blob)
+
+    def test_round_trip_random(self):
+        for _ in range(1000):
+            n = randint(1, PTRSIZE - 1);
+            blob = os.urandom(n)
+            i = _read_n(iter(blob), n)
+            self.assertEqual(_write_n(n, i), blob)
+
+    def test_read_n_untouch(self):
+        it = iter(b"\x00XY")
+        self.assertEqual(_read_n(it, 1), 0)
+        self.assertEqual(next(it), ord('X'))
+        self.assertEqual(_read_n(it, 0), 0)
+        self.assertEqual(next(it), ord('Y'))
+        self.assertRaises(StopIteration, _read_n, it, 1)
+
+    def test_read_n_item_errors(self):
+        for v in -1, 256:
+            self.assertRaises(ValueError, _read_n, iter([3, v]), 2)
+
+        for v in None, "F", Ellipsis, []:
+            self.assertRaises(TypeError, _read_n, iter([3, v]), 2)
+
+    def test_read_n_negative(self):
+        it = iter(PTRSIZE * b"\xff")
+        self.assertRaisesMessage(
+            ValueError,
+            "read %d bytes got negative value: -1" % PTRSIZE,
+            _read_n, it, PTRSIZE)
+
 # ---------------------------------------------------------------------------
 
 if __name__ == '__main__':
